const coral = @import("coral");

const tokens = @import("./tokens.zig");

const types = @import("./types.zig");

pub const BinaryOperation = enum {
	addition,
	subtraction,
	multiplication,
	division
};

pub const ParsedExpression = union (enum) {
	valid: Expression,
	invalid: []const u8,

	pub fn init(allocator: coral.io.Allocator, tokenizer: *tokens.Tokenizer) coral.io.AllocationError!ParsedExpression {
		var parsed_term_expression = try init_term(allocator, tokenizer);

		switch (parsed_term_expression) {
			.valid => |*term_expression| {
				var expression = term_expression.*;
				var is_invalid = true;

				defer if (is_invalid) {
					expression.deinit(allocator);
				};

				if (tokenizer.current_token == .symbol_plus) {
					if (!tokenizer.step()) {
						return ParsedExpression{.invalid = "expected right-hand side of expression after `+`"};
					}

					var parsed_binary_expression = try init_binary(allocator, tokenizer, &expression, .addition);

					expression = switch (parsed_binary_expression) {
						.valid => |binary_expression| binary_expression,
						.invalid => |details| return ParsedExpression{.invalid = details},
					};
				}

				if (tokenizer.current_token == .symbol_minus) {
					if (!tokenizer.step()) {
						return ParsedExpression{.invalid = "expected right-hand side of expression after `-`"};
					}

					var parsed_binary_expression = try init_binary(allocator, tokenizer, &expression, .subtraction);

					expression = switch (parsed_binary_expression) {
						.valid => |binary_expression| binary_expression,
						.invalid => |details| return ParsedExpression{.invalid = details},
					};
				}

				is_invalid = false;

				return ParsedExpression{.valid = expression};
			},

			.invalid => |details| {
				return ParsedExpression{.invalid = details};
			},
		}
	}

	fn init_binary(allocator: coral.io.Allocator, tokenizer: *tokens.Tokenizer, lhs_expression: *const Expression, operation: BinaryOperation) coral.io.AllocationError!ParsedExpression {
		var parsed_expression = try init_term(allocator, tokenizer);

		switch (parsed_expression) {
			.valid => |*expression| {
				errdefer expression.deinit(allocator);

				const rhs_expression = try coral.io.allocate_one(allocator, expression.*);

				errdefer coral.io.deallocate(allocator, rhs_expression);

				return ParsedExpression{
					.valid = .{
						.binary_operation = .{
							.kind = operation,
							.lhs_expression = try coral.io.allocate_one(allocator, lhs_expression.*),
							.rhs_expression = rhs_expression,
						},
					},
				};
			},

			.invalid => |details| {
				return ParsedExpression{.invalid = details};
			},
		}
	}

	fn init_factor(allocator: coral.io.Allocator, tokenizer: *tokens.Tokenizer) coral.io.AllocationError!ParsedExpression {
		switch (tokenizer.current_token) {
			.symbol_paren_left => {
				if (!tokenizer.step()) {
					return ParsedExpression{.invalid = "expected an expression after `(`"};
				}

				var parsed_expression = try ParsedExpression.init(allocator, tokenizer);

				switch (parsed_expression) {
					.valid => |*expression| {
						var is_invalid = true;

						defer if (is_invalid) {
							expression.deinit(allocator);
						};

						if ((!tokenizer.step()) or (tokenizer.current_token != .symbol_paren_right)) {
							return ParsedExpression{.invalid = "expected a closing `)` after expression"};
						}

						is_invalid = false;

						return ParsedExpression{
							.valid = .{.grouped_expression = try coral.io.allocate_one(allocator, expression.*)},
						};
					},

					.invalid => |details| {
						return ParsedExpression{.invalid = details};
					}
				}
			},

			.integer => |value| {
				defer _ = tokenizer.step();

				return ParsedExpression{
					.valid = .{
						.integer_literal = coral.utf8.parse_int(
							@typeInfo(types.Integer).Int,
							value, .{}) catch |parse_error| {

							return ParsedExpression{
								.invalid = switch (parse_error) {
									error.BadSyntax => "invalid integer literal",
									error.IntOverflow => "integer literal is too big",
								}
							};
						},
					},
				};
			},

			.real => |value| {
				defer _ = tokenizer.step();

				return ParsedExpression{
					.valid = .{
						.float_literal = coral.utf8.parse_float(
							@typeInfo(types.Float).Float,
							value) catch |parse_error| {

							return ParsedExpression{
								.invalid = switch (parse_error) {
									error.BadSyntax => "invalid float literal",
								},
							};
						},
					},
				};
			},

			.string => |value| {
				defer _ = tokenizer.step();

				return ParsedExpression{
					.valid = .{
						.string_literal = value,
					},
				};
			},

			.symbol_minus => {
				if (!tokenizer.step()) {
					return ParsedExpression{.invalid = "expected expression after numeric negation (`-`)"};
				}

				return try init_unary(allocator, tokenizer, .numeric_negation);
			},

			.symbol_bang => {
				if (!tokenizer.step()) {
					return ParsedExpression{.invalid = "expected expression after boolean negation (`!`)"};
				}

				return try init_unary(allocator, tokenizer, .boolean_negation);
			},

			else => return ParsedExpression{.invalid = "unexpected token in expression"},
		}
	}

	fn init_term(allocator: coral.io.Allocator, tokenizer: *tokens.Tokenizer) coral.io.AllocationError!ParsedExpression {
		var parsed_factor_expression = try init_factor(allocator, tokenizer);

		switch (parsed_factor_expression) {
			.valid => |*factor_expression| {
				var expression = factor_expression.*;
				var is_invalid = true;

				defer if (is_invalid) {
					expression.deinit(allocator);
				};

				if (tokenizer.current_token == .symbol_asterisk) {
					if (!tokenizer.step()) {
						return ParsedExpression{.invalid = "expected right-hand side of expression after `*`"};
					}

					var parsed_binary_expression = try init_binary(allocator, tokenizer, &expression, .multiplication);

					expression = switch (parsed_binary_expression) {
						.valid => |binary_expression| binary_expression,
						.invalid => |details| return ParsedExpression{.invalid = details},
					};
				}

				if (tokenizer.current_token == .symbol_forward_slash) {
					if (!tokenizer.step()) {
						return ParsedExpression{.invalid = "expected right-hand side of expression after `/`"};
					}

					var parsed_binary_expression = try init_binary(allocator, tokenizer, &expression, .division);

					expression = switch (parsed_binary_expression) {
						.valid => |binary_expression| binary_expression,
						.invalid => |details| return ParsedExpression{.invalid = details},
					};
				}

				is_invalid = false;

				return ParsedExpression{.valid = expression};
			},

			.invalid => |details| {
				return ParsedExpression{.invalid = details};
			},
		}
	}

	fn init_unary(allocator: coral.io.Allocator, tokenizer: *tokens.Tokenizer, operation: UnaryOperation) coral.io.AllocationError!ParsedExpression {
		var parsed_factor_expression = try init_factor(allocator, tokenizer);

		switch (parsed_factor_expression) {
			.valid => |*factor_expression| {
				errdefer factor_expression.deinit(allocator);

				return ParsedExpression{
					.valid = .{
						.unary_operation = .{
							.kind = operation,
							.expression = try coral.io.allocate_one(allocator, factor_expression.*),
						},
					},
				};
			},

			.invalid => |details| {
				return ParsedExpression{.invalid = details};
			},
		}
	}
};

pub const ParsedStatements = union (enum) {
	valid: Statements,
	invalid: []const u8,

	pub fn init(allocator: coral.io.Allocator, tokenizer: *tokens.Tokenizer) coral.io.AllocationError!ParsedStatements {
		var statements_list = try Statements.List.init(allocator, 0);
		var has_returned = false;
		var is_invalid = true;

		defer if (is_invalid) {
			for (statements_list.values) |*statement| {
				statement.deinit(allocator);
			}

			statements_list.deinit(allocator);
		};

		while (tokenizer.step()) {
			switch (tokenizer.current_token) {
				.newline => {},

				.keyword_return => {
					if (has_returned) {
						return ParsedStatements{.invalid = "cannot return more than once per function scope"};
					}

					if (tokenizer.step() and (tokenizer.current_token != .newline)) {
						var parsed_expression = try ParsedExpression.init(allocator, tokenizer);

						switch (parsed_expression) {
							.valid => |*expression| {
								errdefer expression.deinit(allocator);

								try statements_list.push_one(allocator, .{
									.return_expression = expression.*,
								});
							},

							.invalid => |details| {
								return ParsedStatements{.invalid = details};
							},
						}
					} else {
						try statements_list.push_one(allocator, .return_nothing);
					}

					if (tokenizer.step() and tokenizer.current_token != .newline) {
						return ParsedStatements{.invalid = "expected newline after expression"};
					}

					has_returned = true;
				},

				else => {
					return ParsedStatements{.invalid = "invalid statement"};
				},
			}
		}

		is_invalid = false;

		return ParsedStatements{
			.valid = .{
				.list = statements_list,
			},
		};
	}
};

pub const Expression = union (enum) {
	nil_literal,
	true_literal,
	false_literal,
	integer_literal: types.Integer,
	float_literal: types.Float,
	string_literal: []const u8,
	table_literal: TableLiteral,
	grouped_expression: *Expression,

	binary_operation: struct {
		kind: BinaryOperation,
		lhs_expression: *Expression,
		rhs_expression: *Expression,
	},

	unary_operation: struct {
		kind: UnaryOperation,
		expression: *Expression,
	},

	const TableLiteral = coral.list.Stack(struct {
		identifier: []const u8,
		expression: *Expression,
	});

	fn deinit(self: *Expression, allocator: coral.io.Allocator) void {
		switch (self.*) {
			.nil_literal, .true_literal, .false_literal, .integer_literal, .float_literal, .string_literal => {},

			.table_literal => |*literal| {
				for (literal.values) |field| {
					field.expression.deinit(allocator);
					coral.io.deallocate(allocator, field.expression);
				}

				literal.deinit(allocator);
			},

			.grouped_expression => |expression| {
				expression.deinit(allocator);
			},

			.binary_operation => |operation| {
				operation.lhs_expression.deinit(allocator);
				coral.io.deallocate(allocator, operation.lhs_expression);
				operation.rhs_expression.deinit(allocator);
				coral.io.deallocate(allocator, operation.rhs_expression);
			},

			.unary_operation => |operation| {
				operation.expression.deinit(allocator);
				coral.io.deallocate(allocator, operation.expression);
			}
		}
	}
};

pub const Statements = struct {
	list: List,

	const List = coral.list.Stack(union (enum) {
		return_expression: Expression,
		return_nothing,

		const Self = @This();

		fn deinit(self: *Self, allocator: coral.io.Allocator) void {
			switch (self.*) {
				.return_expression => |*expression| {
					expression.deinit(allocator);
				},

				.return_nothing => {},
			}
		}
	});

	pub fn deinit(self: *Statements, allocator: coral.io.Allocator) void {
		for (self.list.values) |*statement| {
			statement.deinit(allocator);
		}

		self.list.deinit(allocator);
	}
};

pub const UnaryOperation = enum {
	boolean_negation,
	numeric_negation,
};
