const coral = @import("coral");

pub const Token = union(enum) {
	unknown: u8,
	newline,

	global: []const u8,
	local: []const u8,

	symbol_assign,
	symbol_plus,
	symbol_minus,
	symbol_asterisk,
	symbol_forward_slash,
	symbol_paren_left,
	symbol_paren_right,
	symbol_bang,
	symbol_comma,
	symbol_at,
	symbol_brace_left,
	symbol_brace_right,
	symbol_bracket_left,
	symbol_bracket_right,
	symbol_period,
	symbol_arrow,

	integer: []const u8,
	real: []const u8,
	string: []const u8,

	keyword_nil,
	keyword_false,
	keyword_true,
	keyword_return,
	keyword_self,

	pub fn text(self: Token) []const u8 {
		return switch (self) {
			.unknown => |unknown| @ptrCast([*]const u8, &unknown)[0 .. 1],
			.newline => "newline",
			.identifier_global => |identifier| identifier,
			.identifier_local => |identifier| identifier,

			.assign => "=",
			.plus => "+",
			.minus => "-",
			.asterisk => "*",
			.forward_slash => "/",
			.paren_left => "(",
			.paren_right => ")",
			.bang => "!",
			.comma => ",",
			.at => "@",
			.brace_left => "{",
			.brace_right => "}",
			.bracket_left => "[",
			.bracket_right => "]",
			.period => ".",
			.arrow => "=>",

			.integer => |literal| literal,
			.real => |literal| literal,
			.string => |literal| literal,

			.keyword_nil => "nil",
			.keyword_false => "false",
			.keyword_true => "true",
			.keyword_return => "return",
			.keyword_self => "self",
		};
	}
};

pub const Tokenizer = struct {
	source: []const u8,
	lines_stepped: usize = 1,
	previous_token: Token = .newline,
	current_token: Token = .newline,

	pub fn has_next(self: Tokenizer) bool {
		return self.source.len != 0;
	}

	pub fn step(self: *Tokenizer) bool {
		self.previous_token = self.current_token;

		var cursor = @as(usize, 0);

		defer self.source = self.source[cursor ..];

		while (self.has_next()) switch (self.source[cursor]) {
			'#' => {
				cursor += 1;

				while (self.has_next() and (self.source[cursor] == '\n')) {
					cursor += 1;
				}
			},

			' ', '\t' => cursor += 1,

			'\n' => {
				cursor += 1;
				self.lines_stepped += 1;
				self.current_token = .newline;

				return true;
			},

			'0' ... '9' => {
				const begin = cursor;

				cursor += 1;

				while (self.has_next()) switch (self.source[cursor]) {
					'0' ... '9' => cursor += 1,

					'.' => {
						cursor += 1;

						while (self.has_next()) switch (self.source[cursor]) {
							'0' ... '9' => cursor += 1,
							else => break,
						};

						self.current_token = .{.real = self.source[begin .. cursor]};

						return true;
					},

					else => break,
				};

				self.current_token = .{.integer = self.source[begin .. cursor]};

				return true;
			},

			'A' ... 'Z', 'a' ... 'z', '_' => {
				const begin = cursor;

				cursor += 1;

				while (cursor < self.source.len) switch (self.source[cursor]) {
					'0'...'9', 'A'...'Z', 'a'...'z', '_' => cursor += 1,
					else => break,
				};

				const identifier = self.source[begin .. cursor];

				coral.debug.assert(identifier.len != 0);

				switch (identifier[0]) {
					'n' => if (coral.io.ends_with(identifier, "il")) {
						self.current_token = .keyword_nil;
					},

					'f' => if (coral.io.ends_with(identifier, "alse")) {
						self.current_token = .keyword_false;
					},

					't' => if (coral.io.ends_with(identifier, "rue")) {
						self.current_token = .keyword_true;
					},

					'r' => if (coral.io.ends_with(identifier, "eturn")) {
						self.current_token = .keyword_return;
					},

					's' => if (coral.io.ends_with(identifier, "elf")) {
						self.current_token = .keyword_self;
					},

					else => self.current_token = .{.local = identifier},
				}

				return true;
			},

			'@' => {
				cursor += 1;

				if (self.has_next()) switch (self.source[cursor]) {
					'A'...'Z', 'a'...'z', '_' => {
						const begin = cursor;

						cursor += 1;

						while (self.has_next()) switch (self.source[cursor]) {
							'0'...'9', 'A'...'Z', 'a'...'z', '_' => cursor += 1,
							else => break,
						};

						self.current_token = .{.global = self.source[begin .. cursor]};

						return true;
					},

					'"' => {
						cursor += 1;

						const begin = cursor;

						cursor += 1;

						while (self.has_next()) switch (self.source[cursor]) {
							'"' => break,
							else => cursor += 1,
						};

						self.current_token = .{.global = self.source[begin .. cursor]};
						cursor += 1;

						return true;
					},

					else => {},
				};

				self.current_token = .symbol_at;

				return true;
			},

			'"' => {
				cursor += 1;

				const begin = cursor;

				cursor += 1;

				while (self.has_next()) switch (self.source[cursor]) {
					'"' => break,
					else => cursor += 1,
				};

				self.current_token = .{.string = self.source[begin .. cursor]};
				cursor += 1;

				return true;
			},

			'{' => {
				self.current_token = .symbol_brace_left;
				cursor += 1;

				return true;
			},

			'}' => {
				self.current_token = .symbol_brace_right;
				cursor += 1;

				return true;
			},

			',' => {
				self.current_token = .symbol_comma;
				cursor += 1;

				return true;
			},

			'!' => {
				self.current_token = .symbol_bang;
				cursor += 1;

				return true;
			},

			')' => {
				self.current_token = .symbol_paren_right;
				cursor += 1;

				return true;
			},

			'(' => {
				self.current_token = .symbol_paren_left;
				cursor += 1;

				return true;
			},

			'/' => {
				self.current_token = .symbol_forward_slash;
				cursor += 1;

				return true;
			},

			'*' => {
				self.current_token = .symbol_asterisk;
				cursor += 1;

				return true;
			},

			'-' => {
				self.current_token = .symbol_minus;
				cursor += 1;

				return true;
			},

			'+' => {
				self.current_token = .symbol_plus;
				cursor += 1;

				return true;
			},

			'=' => {
				self.current_token = .symbol_assign;
				cursor += 1;

				return true;
			},

			'.' => {
				self.current_token = .symbol_period;
				cursor += 1;

				return true;
			},

			else => {
				self.current_token = .{.unknown = self.source[cursor]};
				cursor += 1;

				return true;
			},
		};

		self.current_token = .newline;

		return false;
	}
};
